import functools
import hashlib
from abc import ABC, abstractmethod, abstractproperty
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Dict,
    FrozenSet,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import pendulum

import dagster._check as check
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.metadata import MetadataMapping, MetadataValue
from dagster._core.definitions.partition import AllPartitionsSubset
from dagster._serdes.serdes import PackableValue, whitelist_for_serdes

from ..asset_subset import AssetSubset, ValidAssetSubset

if TYPE_CHECKING:
    from ..auto_materialize_rule import AutoMaterializeRule
    from .asset_condition_evaluation_context import AssetConditionEvaluationContext


T = TypeVar("T")


@whitelist_for_serdes
class HistoricalAllPartitionsSubsetSentinel(NamedTuple):
    """Serializable indicator that this value was an AllPartitionsSubset at serialization time, but
    the partitions may have changed since that time.
    """


@whitelist_for_serdes
class AssetConditionSnapshot(NamedTuple):
    """A serializable snapshot of a node in the AutomationCondition tree."""

    class_name: str
    description: str
    unique_id: str


@whitelist_for_serdes
class AssetSubsetWithMetadata(NamedTuple):
    """An asset subset with metadata that corresponds to it."""

    subset: AssetSubset
    metadata: MetadataMapping

    @property
    def frozen_metadata(self) -> FrozenSet[Tuple[str, MetadataValue]]:
        return frozenset(self.metadata.items())


def get_serializable_candidate_subset(
    candidate_subset: Union[AssetSubset, HistoricalAllPartitionsSubsetSentinel],
) -> Union[AssetSubset, HistoricalAllPartitionsSubsetSentinel]:
    """Do not serialize the candidate subset directly if it is an AllPartitionsSubset."""
    if isinstance(candidate_subset, AssetSubset) and isinstance(
        candidate_subset.value, AllPartitionsSubset
    ):
        return HistoricalAllPartitionsSubsetSentinel()
    return candidate_subset


class AssetConditionResult(NamedTuple):
    condition: "AssetCondition"
    start_timestamp: float
    end_timestamp: float

    true_subset: AssetSubset
    candidate_subset: AssetSubset
    subsets_with_metadata: Sequence[AssetSubsetWithMetadata]
    extra_state: PackableValue

    child_results: Sequence["AssetConditionResult"]

    @staticmethod
    def create_from_children(
        context: "AssetConditionEvaluationContext",
        true_subset: AssetSubset,
        child_results: Sequence["AssetConditionResult"],
    ) -> "AssetConditionResult":
        """Returns a new AssetConditionEvaluation from the given child results."""
        return AssetConditionResult(
            condition=context.condition,
            start_timestamp=context.start_timestamp,
            end_timestamp=pendulum.now("UTC").timestamp(),
            true_subset=true_subset,
            candidate_subset=context.candidate_subset,
            subsets_with_metadata=[],
            child_results=child_results,
            extra_state=None,
        )

    @staticmethod
    def create(
        context: "AssetConditionEvaluationContext",
        true_subset: AssetSubset,
        subsets_with_metadata: Sequence[AssetSubsetWithMetadata] = [],
        extra_state: PackableValue = None,
    ) -> "AssetConditionResult":
        """Returns a new AssetConditionEvaluation from the given parameters."""
        return AssetConditionResult(
            condition=context.condition,
            start_timestamp=context.start_timestamp,
            end_timestamp=pendulum.now("UTC").timestamp(),
            true_subset=true_subset,
            candidate_subset=context.candidate_subset,
            subsets_with_metadata=subsets_with_metadata,
            child_results=[],
            extra_state=extra_state,
        )


@whitelist_for_serdes
class AssetConditionEvaluation(NamedTuple):
    """Serializable representation of the results of evaluating a node in the evaluation tree."""

    condition_snapshot: AssetConditionSnapshot
    start_timestamp: Optional[float]
    end_timestamp: Optional[float]

    true_subset: AssetSubset
    candidate_subset: Union[AssetSubset, HistoricalAllPartitionsSubsetSentinel]
    subsets_with_metadata: Sequence[AssetSubsetWithMetadata]

    child_evaluations: Sequence["AssetConditionEvaluation"]

    @property
    def asset_key(self) -> AssetKey:
        return self.true_subset.asset_key

    @staticmethod
    def from_result(result: AssetConditionResult) -> "AssetConditionEvaluation":
        return AssetConditionEvaluation(
            condition_snapshot=result.condition.snapshot,
            start_timestamp=result.start_timestamp,
            end_timestamp=result.end_timestamp,
            true_subset=result.true_subset,
            candidate_subset=get_serializable_candidate_subset(result.candidate_subset),
            subsets_with_metadata=result.subsets_with_metadata,
            child_evaluations=[
                AssetConditionEvaluation.from_result(child_result)
                for child_result in result.child_results
            ],
        )

    def equivalent_to_stored_evaluation(self, other: Optional["AssetConditionEvaluation"]) -> bool:
        """Returns if this evaluation is functionally equivalent to the given stored evaluation.
        This is used to determine if it is necessary to store this new evaluation in the database.
        Noteably, the timestamps on successive evaluations will always be different, so a simple
        equality check would be too strict.
        """
        return (
            other is not None
            and self.condition_snapshot == other.condition_snapshot
            and self.true_subset == other.true_subset
            # the candidate subset gets modified during serialization
            and get_serializable_candidate_subset(self.candidate_subset)
            == get_serializable_candidate_subset(other.candidate_subset)
            and self.subsets_with_metadata == other.subsets_with_metadata
            and len(self.child_evaluations) == len(other.child_evaluations)
            and all(
                self_child.equivalent_to_stored_evaluation(other_child)
                for self_child, other_child in zip(self.child_evaluations, other.child_evaluations)
            )
        )

    def discarded_subset(self) -> Optional[AssetSubset]:
        """Returns the AssetSubset representing asset partitions that were discarded during this
        evaluation. Note that 'discarding' is a deprecated concept that is only used for backwards
        compatibility.
        """
        if len(self.child_evaluations) != 3:
            return None
        not_discard_evaluation = self.child_evaluations[-1]
        discard_evaluation = not_discard_evaluation.child_evaluations[0]
        return discard_evaluation.true_subset

    def get_requested_or_discarded_subset(self) -> AssetSubset:
        discarded_subset = self.discarded_subset()
        if discarded_subset is None:
            return self.true_subset
        else:
            # the true_subset and discarded_subset were created on the same tick, so they are
            # guaranteed to be compatible with each other
            return ValidAssetSubset(*self.true_subset) | discarded_subset

    def for_child(self, child_condition: "AssetCondition") -> Optional["AssetConditionEvaluation"]:
        """Returns the evaluation of a given child condition by finding the child evaluation that
        has an identical hash to the given condition.
        """
        child_unique_id = child_condition.snapshot.unique_id
        for child_evaluation in self.child_evaluations:
            if child_evaluation.condition_snapshot.unique_id == child_unique_id:
                return child_evaluation

        return None

    def with_run_ids(self, run_ids: AbstractSet[str]) -> "AssetConditionEvaluationWithRunIds":
        return AssetConditionEvaluationWithRunIds(evaluation=self, run_ids=frozenset(run_ids))

    def legacy_num_skipped(self) -> int:
        if len(self.child_evaluations) < 2:
            return 0

        not_skip_evaluation = self.child_evaluations[-1]
        skip_evaluation = not_skip_evaluation.child_evaluations[0]
        return skip_evaluation.true_subset.size - self.legacy_num_discarded()

    def legacy_num_discarded(self) -> int:
        discarded_subset = self.discarded_subset()
        if discarded_subset is None:
            return 0
        return discarded_subset.size


@whitelist_for_serdes
class AssetConditionEvaluationState(NamedTuple):
    """Incremental state calculated during the evaluation of an AssetCondition. This may be used
    on the subsequent evaluation to make the computation more efficient.

    Attributes:
        previous_evaluation: The computed AssetConditionEvaluation.
        previous_tick_evaluation_timestamp: The evaluation_timestamp at which the evaluation was performed.
        max_storage_id: The maximum storage ID over all events used in this computation.
        extra_state_by_unique_id: A mapping from the unique ID of each condition in the evaluation
            tree to the extra state that was calculated for it, if any.
    """

    previous_evaluation: AssetConditionEvaluation
    previous_tick_evaluation_timestamp: Optional[float]

    max_storage_id: Optional[int]
    extra_state_by_unique_id: Mapping[str, PackableValue]

    @property
    def asset_key(self) -> AssetKey:
        return self.previous_evaluation.asset_key

    @property
    def true_subset(self) -> AssetSubset:
        return self.previous_evaluation.true_subset

    @staticmethod
    def create(
        context: "AssetConditionEvaluationContext", root_result: AssetConditionResult
    ) -> "AssetConditionEvaluationState":
        """Convenience constructor to generate an AssetConditionEvaluationState from the root result
        and the context in which it was evaluated.
        """

        # flatten the extra state into a single dict
        def _flatten_extra_state(r: AssetConditionResult) -> Mapping[str, PackableValue]:
            extra_state: Dict[str, PackableValue] = (
                {r.condition.unique_id: r.extra_state} if r.extra_state else {}
            )
            for child in r.child_results:
                extra_state.update(_flatten_extra_state(child))
            return extra_state

        return AssetConditionEvaluationState(
            previous_evaluation=AssetConditionEvaluation.from_result(root_result),
            previous_tick_evaluation_timestamp=context.evaluation_time.timestamp(),
            max_storage_id=context.new_max_storage_id,
            extra_state_by_unique_id=_flatten_extra_state(root_result),
        )

    def get_extra_state(self, condition: "AssetCondition", as_type: Type[T]) -> Optional[T]:
        """Returns the value from the extras dict for the given condition, if it exists and is of
        the expected type. Otherwise, returns None.
        """
        extra_state = self.extra_state_by_unique_id.get(condition.unique_id)
        if isinstance(extra_state, as_type):
            return extra_state
        return None


@whitelist_for_serdes
class AssetConditionEvaluationWithRunIds(NamedTuple):
    """A union of an AssetConditionEvaluation and the set of run IDs that have been launched in
    response to it.
    """

    evaluation: AssetConditionEvaluation
    run_ids: FrozenSet[str]

    @property
    def asset_key(self) -> AssetKey:
        return self.evaluation.asset_key

    @property
    def num_requested(self) -> int:
        return self.evaluation.true_subset.size


class AssetCondition(ABC):
    """An AssetCondition represents some state of the world that can influence if an asset
    partition should be materialized or not. AutomationConditions can be combined together to create
    new conditions using the `&` (and), `|` (or), and `~` (not) operators.
    """

    @property
    def unique_id(self) -> str:
        parts = [
            self.__class__.__name__,
            *[child.unique_id for child in self.children],
        ]
        return hashlib.md5("".join(parts).encode()).hexdigest()

    @abstractmethod
    def evaluate(self, context: "AssetConditionEvaluationContext") -> AssetConditionResult:
        raise NotImplementedError()

    @abstractproperty
    def description(self) -> str:
        raise NotImplementedError()

    def __and__(self, other: "AssetCondition") -> "AssetCondition":
        # group AndAutomationConditions together
        if isinstance(self, AndAssetCondition):
            return AndAssetCondition(children=[*self.children, other])
        return AndAssetCondition(children=[self, other])

    def __or__(self, other: "AssetCondition") -> "AssetCondition":
        # group OrAutomationConditions together
        if isinstance(self, OrAssetCondition):
            return OrAssetCondition(children=[*self.children, other])
        return OrAssetCondition(children=[self, other])

    def __invert__(self) -> "AssetCondition":
        return NotAssetCondition(children=[self])

    @property
    def is_legacy(self) -> bool:
        """Returns if this condition is in the legacy format. This is used to determine if we can
        do certain types of backwards-compatible operations on it.
        """
        return (
            isinstance(self, AndAssetCondition)
            and len(self.children) in {2, 3}
            and isinstance(self.children[0], OrAssetCondition)
            and isinstance(self.children[1], NotAssetCondition)
            # the third child is the discard condition, which is optional
            and (len(self.children) == 2 or isinstance(self.children[2], NotAssetCondition))
        )

    @property
    def children(self) -> Sequence["AssetCondition"]:
        return []

    @property
    def not_skip_condition(self) -> Optional["AssetCondition"]:
        if not self.is_legacy:
            return None
        return self.children[1]

    @property
    def not_discard_condition(self) -> Optional["AssetCondition"]:
        if not self.is_legacy or not len(self.children) == 3:
            return None
        return self.children[-1]

    @functools.cached_property
    def snapshot(self) -> AssetConditionSnapshot:
        """Returns a snapshot of this condition that can be used for serialization."""
        return AssetConditionSnapshot(
            class_name=self.__class__.__name__,
            description=self.description,
            unique_id=self.unique_id,
        )

    @staticmethod
    def parent_newer() -> "AssetCondition":
        """Returns an AssetCondition that is true for an asset partition when at least one parent
        asset partition is newer than it.
        """
        from ..auto_materialize_rule import AutoMaterializeRule

        return RuleCondition(AutoMaterializeRule.materialize_on_parent_updated())

    @staticmethod
    def missing() -> "AssetCondition":
        """Returns an AssetCondition that is true for an asset partition when it has never been
        materialized.
        """
        from ..auto_materialize_rule import AutoMaterializeRule

        return RuleCondition(AutoMaterializeRule.materialize_on_missing())

    @staticmethod
    def parent_missing() -> "AssetCondition":
        """Returns an AssetCondition that is true for an asset partition when at least one parent
        asset partition has never been materialized or observed.
        """
        from ..auto_materialize_rule import AutoMaterializeRule

        return RuleCondition(AutoMaterializeRule.skip_on_parent_missing())

    @staticmethod
    def updated_since_cron(cron_schedule: str, timezone: str = "UTC") -> "AssetCondition":
        """Returns an AssetCondition that is true for an asset partition when it has been updated
        since the latest tick of the given cron schedule. For partitioned assets with a time
        component, this can only be true for the the most recent partition.
        """
        from ..auto_materialize_rule import AutoMaterializeRule

        return ~RuleCondition(AutoMaterializeRule.materialize_on_cron(cron_schedule, timezone))

    @staticmethod
    def parents_updated_since_cron(cron_schedule: str, timezone: str = "UTC") -> "AssetCondition":
        """Returns an AssetCondition that is true for an asset partition when all parent asset
        partitions have been updated more recently than the latest tick of the given cron schedule.
        """
        from ..auto_materialize_rule import AutoMaterializeRule

        return ~RuleCondition(
            AutoMaterializeRule.skip_on_not_all_parents_updated_since_cron(cron_schedule, timezone)
        )


class RuleCondition(
    NamedTuple("_RuleCondition", [("rule", "AutoMaterializeRule")]),
    AssetCondition,
):
    """This class represents the condition that a particular AutoMaterializeRule is satisfied."""

    @property
    def unique_id(self) -> str:
        parts = [self.rule.__class__.__name__, self.description]
        return hashlib.md5("".join(parts).encode()).hexdigest()

    @property
    def description(self) -> str:
        return self.rule.description

    def evaluate(self, context: "AssetConditionEvaluationContext") -> AssetConditionResult:
        context.root_context.daemon_context._verbose_log_fn(  # noqa
            f"Evaluating rule: {self.rule.to_snapshot()}"
        )
        evaluation_result = self.rule.evaluate_for_asset(context)
        context.root_context.daemon_context._verbose_log_fn(  # noqa
            f"Rule returned {evaluation_result.true_subset.size} partitions "
            f"({evaluation_result.end_timestamp - evaluation_result.start_timestamp:.2f} seconds)"
        )
        return evaluation_result


class AndAssetCondition(
    NamedTuple("_AndAssetCondition", [("children", Sequence[AssetCondition])]),
    AssetCondition,
):
    """This class represents the condition that all of its children evaluate to true."""

    @property
    def description(self) -> str:
        return "All of"

    def evaluate(self, context: "AssetConditionEvaluationContext") -> AssetConditionResult:
        child_results: List[AssetConditionResult] = []
        true_subset = context.candidate_subset
        for child in self.children:
            child_context = context.for_child(condition=child, candidate_subset=true_subset)
            child_result = child.evaluate(child_context)
            child_results.append(child_result)
            true_subset &= child_result.true_subset
        return AssetConditionResult.create_from_children(context, true_subset, child_results)


class OrAssetCondition(
    NamedTuple("_OrAssetCondition", [("children", Sequence[AssetCondition])]),
    AssetCondition,
):
    """This class represents the condition that any of its children evaluate to true."""

    @property
    def description(self) -> str:
        return "Any of"

    def evaluate(self, context: "AssetConditionEvaluationContext") -> AssetConditionResult:
        child_results: List[AssetConditionResult] = []
        true_subset = context.empty_subset()
        for child in self.children:
            child_context = context.for_child(
                condition=child, candidate_subset=context.candidate_subset
            )
            child_result = child.evaluate(child_context)
            child_results.append(child_result)
            true_subset |= child_result.true_subset
        return AssetConditionResult.create_from_children(context, true_subset, child_results)


class NotAssetCondition(
    NamedTuple("_NotAssetCondition", [("children", Sequence[AssetCondition])]),
    AssetCondition,
):
    """This class represents the condition that none of its children evaluate to true."""

    def __new__(cls, children: Sequence[AssetCondition]):
        check.invariant(len(children) == 1)
        return super().__new__(cls, children)

    @property
    def description(self) -> str:
        return "Not"

    @property
    def child(self) -> AssetCondition:
        return self.children[0]

    def evaluate(self, context: "AssetConditionEvaluationContext") -> AssetConditionResult:
        child_context = context.for_child(
            condition=self.child, candidate_subset=context.candidate_subset
        )
        child_result = self.child.evaluate(child_context)
        true_subset = context.candidate_subset - child_result.true_subset

        return AssetConditionResult.create_from_children(context, true_subset, [child_result])



from typing import AbstractSet, Iterable, List, Mapping, Optional, Sequence, Union

from dagster._core.definitions.asset_check_spec import AssetCheckKey
from dagster._core.definitions.asset_checks import AssetChecksDefinition
from dagster._core.definitions.asset_graph import AssetGraph, AssetKeyOrCheckKey
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.auto_materialize_policy import AutoMaterializePolicy
from dagster._core.definitions.backfill_policy import BackfillPolicy
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.freshness_policy import FreshnessPolicy
from dagster._core.definitions.partition import PartitionsDefinition
from dagster._core.definitions.partition_mapping import PartitionMapping
from dagster._core.definitions.source_asset import SourceAsset
from dagster._core.selector.subset_selector import DependencyGraph, generate_asset_dep_graph
from dagster._utils.cached_method import cached_method


class InternalAssetGraph(AssetGraph):
    def __init__(
        self,
        assets_defs: Sequence[AssetsDefinition],
        source_assets: Sequence[SourceAsset],
        asset_checks_defs: Sequence[AssetChecksDefinition],
    ):
        from dagster._core.definitions.external_asset import create_external_asset_from_source_asset

        self._assets_defs = [
            *assets_defs,
            *(create_external_asset_from_source_asset(sa) for sa in source_assets),
        ]
        self._assets_defs_by_key = {key: asset for asset in self._assets_defs for key in asset.keys}
        self._assets_defs_by_check_key = {
            **{check_key: asset for asset in assets_defs for check_key in asset.check_keys},
            **{
                check_key: self._assets_defs_by_key[check.asset_key]
                for check in asset_checks_defs
                for check_key in check.keys
                if check.asset_key in self._assets_defs_by_key
            },
        }

        self._asset_checks_defs = asset_checks_defs
        self._asset_checks_defs_by_key = {
            key: check for check in asset_checks_defs for key in check.keys
        }

    @staticmethod
    def from_assets(
        all_assets: Iterable[Union[AssetsDefinition, SourceAsset]],
        asset_checks: Optional[Sequence[AssetChecksDefinition]] = None,
    ) -> "InternalAssetGraph":
        from dagster._core.definitions.internal_asset_graph import InternalAssetGraph

        assets_defs: List[AssetsDefinition] = []
        source_assets: List[SourceAsset] = []
        for asset in all_assets:
            if isinstance(asset, SourceAsset):
                source_assets.append(asset)
            else:  # AssetsDefinition
                assets_defs.append(asset)
        return InternalAssetGraph(
            assets_defs=assets_defs,
            asset_checks_defs=asset_checks or [],
            source_assets=source_assets,
        )

    @property
    def asset_check_keys(self) -> AbstractSet[AssetCheckKey]:
        return {
            *(key for check in self._asset_checks_defs for key in check.keys),
            *(key for asset in self._assets_defs for key in asset.check_keys),
        }

    @property
    def assets_defs(self) -> Sequence[AssetsDefinition]:
        return self._assets_defs

    def get_assets_def(self, asset_key: AssetKey) -> AssetsDefinition:
        return self._assets_defs_by_key[asset_key]

    def has_asset(self, asset_key: AssetKey) -> bool:
        return asset_key in self._assets_defs_by_key

    def get_assets_def_for_check(
        self, asset_check_key: AssetCheckKey
    ) -> Optional[AssetsDefinition]:
        return self._assets_defs_by_check_key.get(asset_check_key)

    @property
    def asset_checks_defs(self) -> Sequence[AssetChecksDefinition]:
        return self._asset_checks_defs

    def get_asset_checks_def(self, asset_check_key: AssetCheckKey) -> AssetChecksDefinition:
        return self._asset_checks_defs_by_key[asset_check_key]

    def has_asset_check(self, asset_check_key: AssetCheckKey) -> bool:
        return asset_check_key in self._asset_checks_defs_by_key

    def includes_materializable_and_external_assets(
        self, asset_keys: AbstractSet[AssetKey]
    ) -> bool:
        """Returns true if the given asset keys contains at least one materializable asset and
        at least one external asset.
        """
        selected_external_assets = self.external_asset_keys & asset_keys
        selected_materializable_assets = self.materializable_asset_keys & asset_keys
        return len(selected_external_assets) > 0 and len(selected_materializable_assets) > 0

    @property
    @cached_method
    def asset_dep_graph(self) -> DependencyGraph[AssetKey]:
        return generate_asset_dep_graph(self._assets_defs, [])

    @property
    @cached_method
    def all_asset_keys(self) -> AbstractSet[AssetKey]:
        return {key for ad in self._assets_defs for key in ad.keys}

    @property
    @cached_method
    def materializable_asset_keys(self) -> AbstractSet[AssetKey]:
        return {key for ad in self._assets_defs if ad.is_materializable for key in ad.keys}

    def is_materializable(self, asset_key: AssetKey) -> bool:
        # Performing an existence check temporarily until we change callsites
        return self.has_asset(asset_key) and self.get_assets_def(asset_key).is_materializable

    @property
    @cached_method
    def observable_asset_keys(self) -> AbstractSet[AssetKey]:
        return {key for ad in self._assets_defs if ad.is_observable for key in ad.keys}

    def is_observable(self, asset_key: AssetKey) -> bool:
        return self.get_assets_def(asset_key).is_observable

    @property
    @cached_method
    def external_asset_keys(self) -> AbstractSet[AssetKey]:
        return {key for ad in self._assets_defs if ad.is_external for key in ad.keys}

    def is_external(self, asset_key: AssetKey) -> bool:
        return self.get_assets_def(asset_key).is_external

    @property
    @cached_method
    def executable_asset_keys(self) -> AbstractSet[AssetKey]:
        return {key for ad in self._assets_defs if ad.is_executable for key in ad.keys}

    def is_executable(self, asset_key: AssetKey) -> bool:
        # Performing an existence check temporarily until we change callsites
        return self.has_asset(asset_key) and self.get_assets_def(asset_key).is_executable

    def asset_keys_for_group(self, group_name: str) -> AbstractSet[AssetKey]:
        return {
            key
            for ad in self._assets_defs
            for key in ad.keys
            if ad.group_names_by_key[key] == group_name
        }

    def get_required_multi_asset_keys(self, asset_key: AssetKey) -> AbstractSet[AssetKey]:
        asset = self.get_assets_def(asset_key)
        return set() if len(asset.keys) <= 1 or asset.can_subset else asset.keys

    def get_required_asset_and_check_keys(
        self, asset_or_check_key: AssetKeyOrCheckKey
    ) -> AbstractSet[AssetKeyOrCheckKey]:
        if isinstance(asset_or_check_key, AssetKey):
            asset = self.get_assets_def(asset_or_check_key)
        else:  # AssetCheckKey
            # only checks emitted by AssetsDefinition have required keys
            if self.has_asset_check(asset_or_check_key):
                return set()
            else:
                asset = self.get_assets_def_for_check(asset_or_check_key)
                if asset is None or asset_or_check_key not in asset.check_keys:
                    return set()
        has_checks = len(asset.check_keys) > 0
        if asset.can_subset or len(asset.keys) <= 1 and not has_checks:
            return set()
        else:
            return {*asset.keys, *asset.check_keys}

    @property
    @cached_method
    def all_group_names(self) -> AbstractSet[str]:
        return {
            group_name for ad in self._assets_defs for group_name in ad.group_names_by_key.values()
        }

    def get_partitions_def(self, asset_key: AssetKey) -> Optional[PartitionsDefinition]:
        # Performing an existence check temporarily until we change callsites
        return self.get_assets_def(asset_key).partitions_def if self.has_asset(asset_key) else None

    def get_partition_mappings(self, asset_key: AssetKey) -> Mapping[AssetKey, PartitionMapping]:
        return self.get_assets_def(asset_key).partition_mappings

    def get_group_name(self, asset_key: AssetKey) -> Optional[str]:
        return self.get_assets_def(asset_key).group_names_by_key.get(asset_key)

    def get_freshness_policy(self, asset_key: AssetKey) -> Optional[FreshnessPolicy]:
        return self.get_assets_def(asset_key).freshness_policies_by_key.get(asset_key)

    def get_auto_materialize_policy(self, asset_key: AssetKey) -> Optional[AutoMaterializePolicy]:
        return self.get_assets_def(asset_key).auto_materialize_policies_by_key.get(asset_key)

    def get_auto_observe_interval_minutes(self, asset_key: AssetKey) -> Optional[float]:
        return self.get_assets_def(asset_key).auto_observe_interval_minutes

    def get_backfill_policy(self, asset_key: AssetKey) -> Optional[BackfillPolicy]:
        return self.get_assets_def(asset_key).backfill_policy

    def get_code_version(self, asset_key: AssetKey) -> Optional[str]:
        return self.get_assets_def(asset_key).code_versions_by_key.get(asset_key)



import datetime
import operator
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Callable,
    NamedTuple,
    Optional,
    Union,
    cast,
)

import dagster._check as check
from dagster._core.definitions.events import AssetKey, AssetKeyPartitionKey
from dagster._core.definitions.partition import (
    AllPartitionsSubset,
    PartitionsDefinition,
    PartitionsSubset,
)
from dagster._core.definitions.time_window_partitions import (
    BaseTimeWindowPartitionsSubset,
)
from dagster._serdes.serdes import (
    NamedTupleSerializer,
    whitelist_for_serdes,
)

if TYPE_CHECKING:
    from dagster._core.instance import DynamicPartitionsStore


class AssetSubsetSerializer(NamedTupleSerializer):
    """Ensures that the inner PartitionsSubset is converted to a serializable form if necessary."""

    def get_storage_name(self) -> str:
        # override this method so all ValidAssetSubsets are serialzied as AssetSubsets
        return "AssetSubset"

    def before_pack(self, value: "AssetSubset") -> "AssetSubset":
        if value.is_partitioned:
            return value._replace(value=value.subset_value.to_serializable_subset())
        return value


@whitelist_for_serdes(serializer=AssetSubsetSerializer)
class AssetSubset(NamedTuple):
    """Represents a set of AssetKeyPartitionKeys for a given AssetKey. For partitioned assets, this
    class uses a PartitionsSubset to represent the set of partitions, enabling lazy evaluation of the
    underlying partition keys. For unpartitioned assets, this class uses a bool to represent whether
    the asset is present or not.
    """

    asset_key: AssetKey
    value: Union[bool, PartitionsSubset]

    @property
    def is_partitioned(self) -> bool:
        return not isinstance(self.value, bool)

    @property
    def bool_value(self) -> bool:
        check.invariant(isinstance(self.value, bool))
        return cast(bool, self.value)

    @property
    def subset_value(self) -> PartitionsSubset:
        check.invariant(isinstance(self.value, PartitionsSubset))
        return cast(PartitionsSubset, self.value)

    @property
    def asset_partitions(self) -> AbstractSet[AssetKeyPartitionKey]:
        if not self.is_partitioned:
            return {AssetKeyPartitionKey(self.asset_key)} if self.bool_value else set()
        else:
            return {
                AssetKeyPartitionKey(self.asset_key, partition_key)
                for partition_key in self.subset_value.get_partition_keys()
            }

    @property
    def size(self) -> int:
        if not self.is_partitioned:
            return int(self.bool_value)
        else:
            return len(self.subset_value)

    def _is_compatible_with_partitions_def(
        self, partitions_def: Optional[PartitionsDefinition]
    ) -> bool:
        if self.is_partitioned:
            # for some PartitionSubset types, we have access to the underlying partitions
            # definitions, so we can ensure those are identical
            if isinstance(self.value, (BaseTimeWindowPartitionsSubset, AllPartitionsSubset)):
                return self.value.partitions_def == partitions_def
            else:
                return partitions_def is not None
        else:
            return partitions_def is None

    def _is_compatible_with_subset(self, other: "AssetSubset") -> bool:
        if isinstance(other.value, (BaseTimeWindowPartitionsSubset, AllPartitionsSubset)):
            return self._is_compatible_with_partitions_def(other.value.partitions_def)
        else:
            return self.is_partitioned == other.is_partitioned

    def as_valid(self, partitions_def: Optional[PartitionsDefinition]) -> "ValidAssetSubset":
        """Converts this AssetSubset to a ValidAssetSubset by returning a copy of this AssetSubset
        if it is compatible with the given PartitionsDefinition, otherwise returns an empty subset.
        """
        if self._is_compatible_with_partitions_def(partitions_def):
            return ValidAssetSubset(*self)
        else:
            return ValidAssetSubset.empty(self.asset_key, partitions_def)

    @staticmethod
    def all(
        asset_key: AssetKey,
        partitions_def: Optional[PartitionsDefinition],
        dynamic_partitions_store: Optional["DynamicPartitionsStore"] = None,
        current_time: Optional[datetime.datetime] = None,
    ) -> "ValidAssetSubset":
        if partitions_def is None:
            return ValidAssetSubset(asset_key=asset_key, value=True)
        else:
            if dynamic_partitions_store is None or current_time is None:
                check.failed(
                    "Must provide dynamic_partitions_store and current_time for partitioned assets."
                )
            return ValidAssetSubset(
                asset_key=asset_key,
                value=AllPartitionsSubset(partitions_def, dynamic_partitions_store, current_time),
            )

    @staticmethod
    def empty(
        asset_key: AssetKey, partitions_def: Optional[PartitionsDefinition]
    ) -> "ValidAssetSubset":
        if partitions_def is None:
            return ValidAssetSubset(asset_key=asset_key, value=False)
        else:
            return ValidAssetSubset(asset_key=asset_key, value=partitions_def.empty_subset())

    @staticmethod
    def from_asset_partitions_set(
        asset_key: AssetKey,
        partitions_def: Optional[PartitionsDefinition],
        asset_partitions_set: AbstractSet[AssetKeyPartitionKey],
    ) -> "ValidAssetSubset":
        if partitions_def is None:
            return ValidAssetSubset(asset_key=asset_key, value=bool(asset_partitions_set))
        else:
            return ValidAssetSubset(
                asset_key=asset_key,
                value=partitions_def.subset_with_partition_keys(
                    {
                        ap.partition_key
                        for ap in asset_partitions_set
                        if ap.partition_key is not None
                    }
                ),
            )

    def __contains__(self, item: AssetKeyPartitionKey) -> bool:
        if not self.is_partitioned:
            return (
                item.asset_key == self.asset_key and item.partition_key is None and self.bool_value
            )
        else:
            return item.asset_key == self.asset_key and item.partition_key in self.subset_value


@whitelist_for_serdes(serializer=AssetSubsetSerializer)
class ValidAssetSubset(AssetSubset):
    """Represents an AssetSubset which is known to be compatible with the current PartitionsDefinition
    of the asset represents.

    This class serializes to a regular AssetSubset, as it is unknown if this value will still be
    valid in the process that deserializes it.

    All operations act over the set of AssetKeyPartitionKeys that the operands represent if the
    subsets are both ValidAssetSubsets. If the other operand cannot be coerced to a ValidAssetSubset,
    it is treated as an empty subset.
    """

    def inverse(
        self,
        partitions_def: Optional[PartitionsDefinition],
        current_time: Optional[datetime.datetime] = None,
        dynamic_partitions_store: Optional["DynamicPartitionsStore"] = None,
    ) -> "ValidAssetSubset":
        """Returns the AssetSubset containing all asset partitions which are not in this AssetSubset."""
        if partitions_def is None:
            return self._replace(value=not self.bool_value)
        else:
            value = partitions_def.subset_with_partition_keys(
                self.subset_value.get_partition_keys_not_in_subset(
                    partitions_def,
                    current_time=current_time,
                    dynamic_partitions_store=dynamic_partitions_store,
                )
            )
            return self._replace(value=value)

    def _oper(self, other: "ValidAssetSubset", oper: Callable[..., Any]) -> "ValidAssetSubset":
        value = oper(self.value, other.value)
        return self._replace(value=value)

    def __sub__(self, other: AssetSubset) -> "ValidAssetSubset":
        """Returns an AssetSubset representing self.asset_partitions - other.asset_partitions."""
        valid_other = self.get_valid(other)
        if not self.is_partitioned:
            return self._replace(value=self.bool_value and not valid_other.bool_value)
        return self._oper(valid_other, operator.sub)

    def __and__(self, other: AssetSubset) -> "ValidAssetSubset":
        """Returns an AssetSubset representing self.asset_partitions & other.asset_partitions."""
        return self._oper(self.get_valid(other), operator.and_)

    def __or__(self, other: AssetSubset) -> "ValidAssetSubset":
        """Returns an AssetSubset representing self.asset_partitions | other.asset_partitions."""
        return self._oper(self.get_valid(other), operator.or_)

    def get_valid(self, other: AssetSubset) -> "ValidAssetSubset":
        """Creates a ValidAssetSubset from the given AssetSubset by returning a copy of the given
        AssetSubset if it is compatible with this AssetSubset, otherwise returns an empty subset.
        """
        if self._is_compatible_with_subset(other):
            return ValidAssetSubset(*other)
        else:
            return self._replace(
                # unfortunately, this is the best way to get an empty partitions subset of an unknown
                # type if you don't have access to the partitions definition
                value=(self.subset_value - self.subset_value) if self.is_partitioned else False
            )



from enum import Enum
from typing import TYPE_CHECKING, Callable, Optional, Sequence, Union

import dagster._check as check
from dagster._core.definitions.external_asset_graph import ExternalAssetGraph
from dagster._core.errors import DagsterInvariantViolationError
from dagster._core.host_representation import ExternalRepository
from dagster._core.workspace.context import BaseWorkspaceRequestContext

if TYPE_CHECKING:
    from dagster._core.definitions.events import (
        AssetKey,
    )


class ChangeReason(Enum):
    NEW = "NEW"
    CODE_VERSION = "CODE_VERSION"
    INPUTS = "INPUTS"


def _get_external_repo_from_context(
    context: BaseWorkspaceRequestContext, code_location_name: str, repository_name: str
) -> Optional[ExternalRepository]:
    """Returns the ExternalRepository specified by the code location name and repository name
    for the provided workspace context. If the repository doesn't exist, return None.
    """
    if context.has_code_location(code_location_name):
        cl = context.get_code_location(code_location_name)
        if cl.has_repository(repository_name):
            return cl.get_repository(repository_name)


class AssetGraphDiffer:
    """Given two asset graphs, base_asset_graph and branch_asset_graph, we can compute how the
    assets in branch_asset_graph have changed with respect to base_asset_graph. The ChangeReason
    enum contains the list of potential changes an asset can undergo. If the base_asset_graph is None,
    this indicates that the branch_asset_graph does not yet exist in the base deployment. In this case
    we will consider every asset New.
    """

    _branch_asset_graph: Optional["ExternalAssetGraph"]
    _branch_asset_graph_load_fn: Optional[Callable[[], "ExternalAssetGraph"]]
    _base_asset_graph: Optional["ExternalAssetGraph"]
    _base_asset_graph_load_fn: Optional[Callable[[], "ExternalAssetGraph"]]

    def __init__(
        self,
        branch_asset_graph: Union["ExternalAssetGraph", Callable[[], "ExternalAssetGraph"]],
        base_asset_graph: Optional[
            Union["ExternalAssetGraph", Callable[[], "ExternalAssetGraph"]]
        ] = None,
    ):
        if base_asset_graph is None:
            # if base_asset_graph is None, then the asset graph in the branch deployment does not exist
            # in the parent deployment
            self._base_asset_graph = None
            self._base_asset_graph_load_fn = None
        elif isinstance(base_asset_graph, ExternalAssetGraph):
            self._base_asset_graph = base_asset_graph
            self._base_asset_graph_load_fn = None
        else:
            self._base_asset_graph = None
            self._base_asset_graph_load_fn = base_asset_graph

        if isinstance(branch_asset_graph, ExternalAssetGraph):
            self._branch_asset_graph = branch_asset_graph
            self._branch_asset_graph_load_fn = None
        else:
            self._branch_asset_graph = None
            self._branch_asset_graph_load_fn = branch_asset_graph

    @classmethod
    def from_external_repositories(
        cls,
        code_location_name: str,
        repository_name: str,
        branch_workspace: BaseWorkspaceRequestContext,
        parent_workspace: BaseWorkspaceRequestContext,
    ) -> "AssetGraphDiffer":
        """Constructs a ParentAssetGraphDiffer for a particular repository in a code location for two
        deployment workspaces, the parent deployment and the branch deployment.

        We cannot make ExternalAssetGraphs directly from the workspaces because if multiple code locations
        use the same asset key, those asset keys will override each other in the dictionaries the ExternalAssetGraph
        creates (see from_repository_handles_and_external_asset_nodes in ExternalAssetGraph). We need to ensure
        that we are comparing assets in the same code location and repository, so we need to make the
        ExternalAssetGraph from an ExternalRepository to ensure that there are no duplicate asset keys
        that could override each other.
        """
        check.inst_param(branch_workspace, "branch_workspace", BaseWorkspaceRequestContext)
        check.inst_param(parent_workspace, "parent_workspace", BaseWorkspaceRequestContext)

        branch_repo = _get_external_repo_from_context(
            branch_workspace, code_location_name, repository_name
        )
        if branch_repo is None:
            raise DagsterInvariantViolationError(
                f"Repository {repository_name} does not exist in code location {code_location_name} for the branch deployment."
            )
        parent_repo = _get_external_repo_from_context(
            parent_workspace, code_location_name, repository_name
        )
        return AssetGraphDiffer(
            branch_asset_graph=lambda: ExternalAssetGraph.from_external_repository(branch_repo),
            base_asset_graph=(lambda: ExternalAssetGraph.from_external_repository(parent_repo))
            if parent_repo is not None
            else None,
        )

    def _compare_parent_and_branch_assets(self, asset_key: "AssetKey") -> Sequence[ChangeReason]:
        """Computes the diff between a branch deployment asset and the
        corresponding parent deployment asset.
        """
        if self.base_asset_graph is None:
            # if the parent asset graph is None, it is because the the asset graph in the branch deployment
            # is new and doesn't exist in the parent deployment. Thus all assets are new.
            return [ChangeReason.NEW]

        if asset_key not in self.base_asset_graph.all_asset_keys:
            return [ChangeReason.NEW]

        changes = []
        if self.branch_asset_graph.get_code_version(
            asset_key
        ) != self.base_asset_graph.get_code_version(asset_key):
            changes.append(ChangeReason.CODE_VERSION)

        if self.branch_asset_graph.get_parents(asset_key) != self.base_asset_graph.get_parents(
            asset_key
        ):
            changes.append(ChangeReason.INPUTS)

        return changes

    def get_changes_for_asset(self, asset_key: "AssetKey") -> Sequence[ChangeReason]:
        """Returns list of ChangeReasons for asset_key as compared to the parent deployment."""
        return self._compare_parent_and_branch_assets(asset_key)

    @property
    def branch_asset_graph(self) -> "ExternalAssetGraph":
        if self._branch_asset_graph is None:
            self._branch_asset_graph = check.not_none(self._branch_asset_graph_load_fn)()
        return self._branch_asset_graph

    @property
    def base_asset_graph(self) -> Optional["ExternalAssetGraph"]:
        if self._base_asset_graph is None and self._base_asset_graph_load_fn is not None:
            self._base_asset_graph = self._base_asset_graph_load_fn()
        return self._base_asset_graph



import logging
import time
from typing import Dict, Iterator, List, NamedTuple, Optional, Sequence, Tuple, cast

import dagster._check as check
from dagster._core.definitions.assets_job import is_base_asset_job_name
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.external_asset_graph import ExternalAssetGraph
from dagster._core.definitions.partition import PartitionsDefinition
from dagster._core.definitions.run_request import RunRequest
from dagster._core.definitions.selector import JobSubsetSelector
from dagster._core.errors import (
    DagsterCodeLocationLoadError,
    DagsterUserCodeUnreachableError,
)
from dagster._core.host_representation import ExternalExecutionPlan, ExternalJob
from dagster._core.instance import DagsterInstance
from dagster._core.snap import ExecutionPlanSnapshot
from dagster._core.storage.dagster_run import DagsterRun, DagsterRunStatus
from dagster._core.workspace.context import BaseWorkspaceRequestContext, IWorkspaceProcessContext
from dagster._utils import SingleInstigatorDebugCrashFlags, check_for_debug_crash, hash_collection

EXECUTION_PLAN_CREATION_RETRIES = 1


class RunRequestExecutionData(NamedTuple):
    external_job: ExternalJob
    external_execution_plan: ExternalExecutionPlan
    partitions_def: Optional[PartitionsDefinition]


def _get_implicit_job_name_for_assets(
    asset_graph: ExternalAssetGraph, asset_keys: Sequence[AssetKey]
) -> Optional[str]:
    job_names = set(asset_graph.get_materialization_job_names(asset_keys[0]))
    for asset_key in asset_keys[1:]:
        job_names &= set(asset_graph.get_materialization_job_names(asset_key))

    return next(job_name for job_name in job_names if is_base_asset_job_name(job_name))


def _execution_plan_targets_asset_selection(
    execution_plan_snapshot: ExecutionPlanSnapshot, asset_selection: Sequence[AssetKey]
) -> bool:
    output_asset_keys = set()
    for step in execution_plan_snapshot.steps:
        if step.key in execution_plan_snapshot.step_keys_to_execute:
            for output in step.outputs:
                asset_key = check.not_none(output.properties).asset_key
                if asset_key:
                    output_asset_keys.add(asset_key)
    return all(key in output_asset_keys for key in asset_selection)


def _get_job_execution_data_from_run_request(
    asset_graph: ExternalAssetGraph,
    run_request: RunRequest,
    instance: DagsterInstance,
    workspace: BaseWorkspaceRequestContext,
    run_request_execution_data_cache: Dict[int, RunRequestExecutionData],
) -> RunRequestExecutionData:
    repo_handle = asset_graph.get_repository_handle(
        cast(Sequence[AssetKey], run_request.asset_selection)[0]
    )
    location_name = repo_handle.code_location_origin.location_name
    job_name = _get_implicit_job_name_for_assets(
        asset_graph, cast(Sequence[AssetKey], run_request.asset_selection)
    )
    if job_name is None:
        check.failed(
            "Could not find an implicit asset job for the given assets:"
            f" {run_request.asset_selection}"
        )

    if not run_request.asset_selection:
        check.failed("Expected RunRequest to have an asset selection")

    pipeline_selector = JobSubsetSelector(
        location_name=location_name,
        repository_name=repo_handle.repository_name,
        job_name=job_name,
        asset_selection=run_request.asset_selection,
        op_selection=None,
    )

    selector_id = hash_collection(pipeline_selector)

    if selector_id not in run_request_execution_data_cache:
        code_location = workspace.get_code_location(repo_handle.code_location_origin.location_name)
        external_job = code_location.get_external_job(pipeline_selector)

        external_execution_plan = code_location.get_external_execution_plan(
            external_job,
            {},
            step_keys_to_execute=None,
            known_state=None,
            instance=instance,
        )

        partitions_def = code_location.get_asset_job_partitions_def(external_job)

        run_request_execution_data_cache[selector_id] = RunRequestExecutionData(
            external_job,
            external_execution_plan,
            partitions_def,
        )

    return run_request_execution_data_cache[selector_id]


def _create_asset_run(
    run_id: Optional[str],
    run_request: RunRequest,
    run_request_index: int,
    instance: DagsterInstance,
    run_request_execution_data_cache: Dict[int, RunRequestExecutionData],
    asset_graph: ExternalAssetGraph,
    workspace_process_context: IWorkspaceProcessContext,
    debug_crash_flags: SingleInstigatorDebugCrashFlags,
    logger: logging.Logger,
) -> DagsterRun:
    """Creates a run on the instance for the given run request. Ensures that the created run results
    in an ExecutionPlan that targets the given asset selection. If it does not, attempts to create
    a valid ExecutionPlan by reloading the workspace.
    """
    from dagster._daemon.controller import RELOAD_WORKSPACE_INTERVAL

    if not run_request.asset_selection:
        check.failed("Expected RunRequest to have an asset selection")

    for _ in range(EXECUTION_PLAN_CREATION_RETRIES + 1):
        # create a new request context for each run in case the code location server
        # is swapped out in the middle of the submission process
        workspace = workspace_process_context.create_request_context()
        execution_data = _get_job_execution_data_from_run_request(
            asset_graph,
            run_request,
            instance,
            workspace=workspace,
            run_request_execution_data_cache=run_request_execution_data_cache,
        )
        check_for_debug_crash(debug_crash_flags, "EXECUTION_PLAN_CREATED")
        check_for_debug_crash(debug_crash_flags, f"EXECUTION_PLAN_CREATED_{run_request_index}")

        # retry until the execution plan targets the asset selection
        if _execution_plan_targets_asset_selection(
            execution_data.external_execution_plan.execution_plan_snapshot,
            check.not_none(run_request.asset_selection),
        ):
            external_job = execution_data.external_job
            external_execution_plan = execution_data.external_execution_plan
            partitions_def = execution_data.partitions_def

            run = instance.create_run(
                job_snapshot=external_job.job_snapshot,
                execution_plan_snapshot=external_execution_plan.execution_plan_snapshot,
                parent_job_snapshot=external_job.parent_job_snapshot,
                job_name=external_job.name,
                run_id=run_id,
                resolved_op_selection=None,
                op_selection=None,
                run_config={},
                step_keys_to_execute=None,
                tags=run_request.tags,
                root_run_id=None,
                parent_run_id=None,
                status=DagsterRunStatus.NOT_STARTED,
                external_job_origin=external_job.get_external_origin(),
                job_code_origin=external_job.get_python_origin(),
                asset_selection=frozenset(run_request.asset_selection),
                asset_check_selection=None,
                asset_job_partitions_def=partitions_def,
            )

            return run

        logger.warning(
            "Execution plan is out of sync with the workspace. Pausing run submission for "
            f"{RELOAD_WORKSPACE_INTERVAL} to allow the execution plan to rebuild with the updated workspace."
        )
        # Sleep for RELOAD_WORKSPACE_INTERVAL seconds since the workspace can be refreshed
        # at most once every interval
        time.sleep(RELOAD_WORKSPACE_INTERVAL)
        # Clear the execution plan cache as this data is no longer valid
        run_request_execution_data_cache = {}

        # If the execution plan does not targets the asset selection, the asset graph
        # likely is outdated and targeting the wrong job, refetch the asset
        # graph from the workspace
        workspace = workspace_process_context.create_request_context()
        asset_graph = ExternalAssetGraph.from_workspace(workspace)

    check.failed(
        f"Failed to target asset selection {run_request.asset_selection} in run after retrying."
    )


def submit_asset_run(
    run_id: Optional[str],
    run_request: RunRequest,
    run_request_index: int,
    instance: DagsterInstance,
    workspace_process_context: IWorkspaceProcessContext,
    asset_graph: ExternalAssetGraph,
    run_request_execution_data_cache: Dict[int, RunRequestExecutionData],
    debug_crash_flags: SingleInstigatorDebugCrashFlags,
    logger: logging.Logger,
) -> DagsterRun:
    """Submits a run for a run request that targets an asset selection. If the run already exists,
    submits the existing run. If the run does not exist, creates a new run and submits it, ensuring
    that the created run targets the given asset selection.
    """
    check.invariant(not run_request.run_config, "Asset run requests have no custom run config")
    asset_keys = check.not_none(run_request.asset_selection)

    check.invariant(len(asset_keys) > 0)

    # check if the run already exists
    existing_run = instance.get_run_by_id(run_id) if run_id else None
    if existing_run:
        if existing_run.status != DagsterRunStatus.NOT_STARTED:
            logger.warn(
                f"Run {run_id} already submitted on a previously interrupted tick, skipping"
            )

            check_for_debug_crash(debug_crash_flags, "RUN_SUBMITTED")
            check_for_debug_crash(debug_crash_flags, f"RUN_SUBMITTED_{run_request_index}")

            return existing_run
        else:
            logger.warn(
                f"Run {run_id} already created on a previously interrupted tick, submitting"
            )
            run_to_submit = existing_run
    else:
        run_to_submit = _create_asset_run(
            run_id,
            run_request,
            run_request_index,
            instance,
            run_request_execution_data_cache,
            asset_graph,
            workspace_process_context,
            debug_crash_flags,
            logger,
        )

    check_for_debug_crash(debug_crash_flags, "RUN_CREATED")
    check_for_debug_crash(debug_crash_flags, f"RUN_CREATED_{run_request_index}")

    instance.submit_run(run_to_submit.run_id, workspace_process_context.create_request_context())

    check_for_debug_crash(debug_crash_flags, "RUN_SUBMITTED")
    check_for_debug_crash(debug_crash_flags, f"RUN_SUBMITTED_{run_request_index}")

    asset_key_str = ", ".join([asset_key.to_user_string() for asset_key in asset_keys])

    logger.info(
        f"Submitted run {run_to_submit.run_id} for assets {asset_key_str} with tags"
        f" {run_request.tags}"
    )

    return run_to_submit


class SubmitRunRequestChunkResult(NamedTuple):
    chunk_submitted_runs: Sequence[Tuple[RunRequest, DagsterRun]]
    retryable_error_raised: bool


def submit_asset_runs_in_chunks(
    run_requests: Sequence[RunRequest],
    reserved_run_ids: Optional[Sequence[str]],
    chunk_size: int,
    instance: DagsterInstance,
    workspace_process_context: IWorkspaceProcessContext,
    asset_graph: ExternalAssetGraph,
    debug_crash_flags: SingleInstigatorDebugCrashFlags,
    logger: logging.Logger,
    backfill_id: Optional[str] = None,
) -> Iterator[Optional[SubmitRunRequestChunkResult]]:
    """Submits runs for a sequence of run requests that target asset selections in chunks. Yields
    None after each run is submitted to allow the daemon to heartbeat, and yields a list of tuples
    of the run request and the submitted run after each chunk is submitted to allow the caller to
    interrupt this process if needed.
    """
    if reserved_run_ids is not None:
        check.invariant(len(run_requests) == len(reserved_run_ids))

    run_request_execution_data_cache = {}
    for chunk_start in range(0, len(run_requests), chunk_size):
        run_request_chunk = run_requests[chunk_start : chunk_start + chunk_size]
        chunk_submitted_runs: List[Tuple[RunRequest, DagsterRun]] = []
        retryable_error_raised = False

        logger.critical(f"{chunk_size}, {chunk_start}, {len(run_request_chunk)}")

        # submit each run in the chunk
        for chunk_idx, run_request in enumerate(run_request_chunk):
            run_request_idx = chunk_start + chunk_idx
            run_id = reserved_run_ids[run_request_idx] if reserved_run_ids else None
            try:
                submitted_run = submit_asset_run(
                    run_id,
                    run_request,
                    run_request_idx,
                    instance,
                    workspace_process_context,
                    asset_graph,
                    run_request_execution_data_cache,
                    debug_crash_flags,
                    logger,
                )
                chunk_submitted_runs.append((run_request, submitted_run))
                # allow the daemon to heartbeat while runs are submitted
                yield None
            except (DagsterUserCodeUnreachableError, DagsterCodeLocationLoadError) as e:
                logger.warning(
                    f"Unable to reach the user code server for assets {run_request.asset_selection}."
                    f" Backfill {backfill_id} will resume execution once the server is available."
                    f"User code server error: {e}"
                )
                retryable_error_raised = True
                # Stop submitting runs if the user code server is unreachable for any
                # given run request
                break

        yield SubmitRunRequestChunkResult(chunk_submitted_runs, retryable_error_raised)



import dataclasses
import json
from dataclasses import dataclass
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
)

from dagster._core.definitions.asset_graph_subset import AssetGraphSubset
from dagster._core.definitions.asset_subset import AssetSubset
from dagster._core.definitions.events import AssetKey
from dagster._serdes.serdes import (
    FieldSerializer,
    JsonSerializableValue,
    PackableValue,
    SerializableNonScalarKeyMapping,
    UnpackContext,
    WhitelistMap,
    pack_value,
    unpack_value,
    whitelist_for_serdes,
)

from .asset_graph import AssetGraph

if TYPE_CHECKING:
    from .asset_condition.asset_condition import (
        AssetConditionEvaluation,
        AssetConditionEvaluationState,
        AssetConditionSnapshot,
    )


@whitelist_for_serdes
class AssetConditionCursorExtras(NamedTuple):
    """Represents additional state that may be optionally saved by an AssetCondition between
    evaluations.
    """

    condition_snapshot: "AssetConditionSnapshot"
    extras: Mapping[str, PackableValue]


class ObserveRequestTimestampSerializer(FieldSerializer):
    def pack(
        self,
        mapping: Mapping[str, float],
        whitelist_map: WhitelistMap,
        descent_path: str,
    ) -> JsonSerializableValue:
        return pack_value(SerializableNonScalarKeyMapping(mapping), whitelist_map, descent_path)

    def unpack(
        self,
        unpacked_value: JsonSerializableValue,
        whitelist_map: WhitelistMap,
        context: UnpackContext,
    ) -> PackableValue:
        return unpack_value(unpacked_value, dict, whitelist_map, context)


@whitelist_for_serdes(
    field_serializers={
        "last_observe_request_timestamp_by_asset_key": ObserveRequestTimestampSerializer
    }
)
@dataclass(frozen=True)
class AssetDaemonCursor:
    """State that's stored between daemon evaluations.

    Attributes:
        evaluation_id (int): The ID of the evaluation that produced this cursor.
        previous_evaluation_state (Sequence[AssetConditionEvaluationInfo]): The evaluation info
            recorded for each asset on the previous tick.
    """

    evaluation_id: int
    previous_evaluation_state: Sequence["AssetConditionEvaluationState"]

    last_observe_request_timestamp_by_asset_key: Mapping[AssetKey, float]

    @staticmethod
    def empty(evaluation_id: int = 0) -> "AssetDaemonCursor":
        return AssetDaemonCursor(
            evaluation_id=evaluation_id,
            previous_evaluation_state=[],
            last_observe_request_timestamp_by_asset_key={},
        )

    @cached_property
    def previous_evaluation_state_by_key(
        self,
    ) -> Mapping[AssetKey, "AssetConditionEvaluationState"]:
        """Efficient lookup of previous evaluation info by asset key."""
        return {
            evaluation_state.asset_key: evaluation_state
            for evaluation_state in self.previous_evaluation_state
        }

    def get_previous_evaluation_state(
        self, asset_key: AssetKey
    ) -> Optional["AssetConditionEvaluationState"]:
        """Returns the AssetConditionCursor associated with the given asset key. If no stored
        cursor exists, returns an empty cursor.
        """
        return self.previous_evaluation_state_by_key.get(asset_key)

    def get_previous_evaluation(self, asset_key: AssetKey) -> Optional["AssetConditionEvaluation"]:
        """Returns the previous AssetConditionEvaluation for a given asset key, if it exists."""
        previous_evaluation_state = self.get_previous_evaluation_state(asset_key)
        return previous_evaluation_state.previous_evaluation if previous_evaluation_state else None

    def with_updates(
        self,
        evaluation_id: int,
        evaluation_timestamp: float,
        newly_observe_requested_asset_keys: Sequence[AssetKey],
        evaluation_state: Sequence["AssetConditionEvaluationState"],
    ) -> "AssetDaemonCursor":
        return dataclasses.replace(
            self,
            evaluation_id=evaluation_id,
            previous_evaluation_state=evaluation_state,
            last_observe_request_timestamp_by_asset_key={
                **self.last_observe_request_timestamp_by_asset_key,
                **{
                    asset_key: evaluation_timestamp
                    for asset_key in newly_observe_requested_asset_keys
                },
            },
        )

    def __hash__(self) -> int:
        return hash(id(self))


# BACKCOMPAT


def get_backcompat_asset_condition_evaluation_state(
    latest_evaluation: "AssetConditionEvaluation",
    latest_storage_id: Optional[int],
    latest_timestamp: Optional[float],
    handled_root_subset: Optional[AssetSubset],
) -> "AssetConditionEvaluationState":
    """Generates an AssetDaemonCursor from information available on the old cursor format."""
    from dagster._core.definitions.asset_condition.asset_condition import (
        AssetConditionEvaluationState,
        RuleCondition,
    )
    from dagster._core.definitions.auto_materialize_rule import MaterializeOnMissingRule

    return AssetConditionEvaluationState(
        previous_evaluation=latest_evaluation,
        previous_tick_evaluation_timestamp=latest_timestamp,
        max_storage_id=latest_storage_id,
        # the only information we need to preserve from the previous cursor is the handled subset
        extra_state_by_unique_id={
            RuleCondition(MaterializeOnMissingRule()).unique_id: handled_root_subset,
        }
        if handled_root_subset and handled_root_subset.size > 0
        else {},
    )


def backcompat_deserialize_asset_daemon_cursor_str(
    cursor_str: str, asset_graph: Optional[AssetGraph], default_evaluation_id: int
) -> AssetDaemonCursor:
    """This serves as a backcompat layer for deserializing the old cursor format. Some information
    is impossible to fully recover, this will recover enough to continue operating as normal.
    """
    from .asset_condition.asset_condition import AssetConditionEvaluation, AssetConditionSnapshot
    from .auto_materialize_rule_evaluation import (
        deserialize_auto_materialize_asset_evaluation_to_asset_condition_evaluation_with_run_ids,
    )

    data = json.loads(cursor_str)

    if isinstance(data, list):
        evaluation_id = data[0] if isinstance(data[0], int) else default_evaluation_id
        return AssetDaemonCursor.empty(evaluation_id)
    elif not isinstance(data, dict):
        return AssetDaemonCursor.empty(default_evaluation_id)
    elif asset_graph is None:
        return AssetDaemonCursor.empty(data.get("evaluation_id", default_evaluation_id))

    serialized_last_observe_request_timestamp_by_asset_key = data.get(
        "last_observe_request_timestamp_by_asset_key", {}
    )
    last_observe_request_timestamp_by_asset_key = {
        AssetKey.from_user_string(key_str): timestamp
        for key_str, timestamp in serialized_last_observe_request_timestamp_by_asset_key.items()
    }

    partition_subsets_by_asset_key = {}
    for key_str, serialized_str in data.get("handled_root_partitions_by_asset_key", {}).items():
        asset_key = AssetKey.from_user_string(key_str)
        partitions_def = asset_graph.get_partitions_def(asset_key) if asset_graph else None
        if not partitions_def:
            continue
        try:
            partition_subsets_by_asset_key[asset_key] = partitions_def.deserialize_subset(
                serialized_str
            )
        except:
            continue

    handled_root_asset_graph_subset = AssetGraphSubset(
        non_partitioned_asset_keys={
            AssetKey.from_user_string(key_str)
            for key_str in data.get("handled_root_asset_keys", set())
        },
        partitions_subsets_by_asset_key=partition_subsets_by_asset_key,
    )

    serialized_latest_evaluation_by_asset_key = data.get("latest_evaluation_by_asset_key", {})
    latest_evaluation_by_asset_key = {}
    for key_str, serialized_evaluation in serialized_latest_evaluation_by_asset_key.items():
        key = AssetKey.from_user_string(key_str)
        partitions_def = asset_graph.get_partitions_def(key) if asset_graph else None

        evaluation = deserialize_auto_materialize_asset_evaluation_to_asset_condition_evaluation_with_run_ids(
            serialized_evaluation, partitions_def
        ).evaluation

        latest_evaluation_by_asset_key[key] = evaluation

    previous_evaluation_state = []
    cursor_keys = (
        asset_graph.materializable_asset_keys
        if asset_graph
        else latest_evaluation_by_asset_key.keys()
    )
    for asset_key in cursor_keys:
        latest_evaluation_result = latest_evaluation_by_asset_key.get(asset_key)
        # create a placeholder evaluation result if we don't have one
        if not latest_evaluation_result:
            partitions_def = asset_graph.get_partitions_def(asset_key) if asset_graph else None
            latest_evaluation_result = AssetConditionEvaluation(
                condition_snapshot=AssetConditionSnapshot("", "", ""),
                true_subset=AssetSubset.empty(asset_key, partitions_def),
                candidate_subset=AssetSubset.empty(asset_key, partitions_def),
                start_timestamp=None,
                end_timestamp=None,
                subsets_with_metadata=[],
                child_evaluations=[],
            )
        backcompat_evaluation_state = get_backcompat_asset_condition_evaluation_state(
            latest_evaluation_result,
            data.get("latest_storage_id"),
            data.get("latest_evaluation_timestamp"),
            handled_root_asset_graph_subset.get_asset_subset(asset_key, asset_graph)
            if asset_graph
            else None,
        )
        previous_evaluation_state.append(backcompat_evaluation_state)

    return AssetDaemonCursor(
        evaluation_id=data.get("evaluation_id") or default_evaluation_id,
        previous_evaluation_state=previous_evaluation_state,
        last_observe_request_timestamp_by_asset_key=last_observe_request_timestamp_by_asset_key,
    )


@whitelist_for_serdes
class LegacyAssetDaemonCursorWrapper(NamedTuple):
    """Wrapper class for the legacy AssetDaemonCursor object, which is not a serializable NamedTuple."""

    serialized_cursor: str

    def get_asset_daemon_cursor(self, asset_graph: Optional[AssetGraph]) -> AssetDaemonCursor:
        return backcompat_deserialize_asset_daemon_cursor_str(
            self.serialized_cursor, asset_graph, 0
        )



import warnings
from collections import defaultdict
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    DefaultDict,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
)

import dagster._check as check
from dagster._core.definitions.asset_check_spec import AssetCheckKey
from dagster._core.definitions.asset_spec import AssetExecutionType
from dagster._core.definitions.assets_job import ASSET_BASE_JOB_PREFIX
from dagster._core.definitions.auto_materialize_policy import AutoMaterializePolicy
from dagster._core.host_representation.external import ExternalRepository
from dagster._core.host_representation.handle import RepositoryHandle
from dagster._core.selector.subset_selector import DependencyGraph
from dagster._core.workspace.workspace import IWorkspace
from dagster._utils.cached_method import cached_method

from .asset_graph import AssetGraph, AssetKeyOrCheckKey
from .backfill_policy import BackfillPolicy
from .events import AssetKey
from .freshness_policy import FreshnessPolicy
from .partition import PartitionsDefinition
from .partition_mapping import PartitionMapping

if TYPE_CHECKING:
    from dagster._core.host_representation.external_data import (
        ExternalAssetCheck,
        ExternalAssetNode,
    )


class ExternalAssetGraph(AssetGraph):
    def __init__(
        self,
        asset_nodes_by_key: Mapping[AssetKey, "ExternalAssetNode"],
        asset_checks_by_key: Mapping[AssetCheckKey, "ExternalAssetCheck"],
        repo_handles_by_key: Mapping[AssetKey, RepositoryHandle],
    ):
        self._asset_nodes_by_key = asset_nodes_by_key
        self._asset_checks_by_key = asset_checks_by_key
        self._repo_handles_by_key = repo_handles_by_key

    @classmethod
    def from_workspace(cls, context: IWorkspace) -> "ExternalAssetGraph":
        code_locations = (
            location_entry.code_location
            for location_entry in context.get_workspace_snapshot().values()
            if location_entry.code_location
        )
        repos = (
            repo
            for code_location in code_locations
            for repo in code_location.get_repositories().values()
        )
        repo_handle_external_asset_nodes: Sequence[
            Tuple[RepositoryHandle, "ExternalAssetNode"]
        ] = []
        asset_checks: Sequence["ExternalAssetCheck"] = []

        for repo in repos:
            for external_asset_node in repo.get_external_asset_nodes():
                repo_handle_external_asset_nodes.append((repo.handle, external_asset_node))

            asset_checks.extend(repo.get_external_asset_checks())

        return cls.from_repository_handles_and_external_asset_nodes(
            repo_handle_external_asset_nodes=repo_handle_external_asset_nodes,
            external_asset_checks=asset_checks,
        )

    @classmethod
    def from_external_repository(
        cls, external_repository: ExternalRepository
    ) -> "ExternalAssetGraph":
        return cls.from_repository_handles_and_external_asset_nodes(
            repo_handle_external_asset_nodes=[
                (external_repository.handle, asset_node)
                for asset_node in external_repository.get_external_asset_nodes()
            ],
            external_asset_checks=external_repository.get_external_asset_checks(),
        )

    @classmethod
    def from_repository_handles_and_external_asset_nodes(
        cls,
        repo_handle_external_asset_nodes: Sequence[Tuple[RepositoryHandle, "ExternalAssetNode"]],
        external_asset_checks: Sequence["ExternalAssetCheck"],
    ) -> "ExternalAssetGraph":
        repo_handles_by_key = {
            node.asset_key: repo_handle
            for repo_handle, node in repo_handle_external_asset_nodes
            if node.is_executable
        }

        # Split the nodes into materializable, observable, and unexecutable nodes. Observable and
        # unexecutable `ExternalAssetNode` represent both source and external assets-- the
        # "External" in "ExternalAssetNode" is unrelated to the "external" in "external asset", this
        # is just an unfortunate naming collision. `ExternalAssetNode` will be renamed eventually.
        materializable_node_pairs: List[Tuple[RepositoryHandle, "ExternalAssetNode"]] = []
        observable_node_pairs: List[Tuple[RepositoryHandle, "ExternalAssetNode"]] = []
        unexecutable_node_pairs: List[Tuple[RepositoryHandle, "ExternalAssetNode"]] = []
        for repo_handle, node in repo_handle_external_asset_nodes:
            if node.is_source and node.is_observable:
                observable_node_pairs.append((repo_handle, node))
            elif node.is_source:
                unexecutable_node_pairs.append((repo_handle, node))
            else:
                materializable_node_pairs.append((repo_handle, node))

        asset_nodes_by_key = {}

        _warn_on_duplicate_nodes(materializable_node_pairs, AssetExecutionType.MATERIALIZATION)
        _warn_on_duplicate_nodes(observable_node_pairs, AssetExecutionType.OBSERVATION)

        # It is possible for multiple nodes to exist that share the same key. This is invalid if
        # more than one node is materializable or if more than one node is observable. It is valid
        # if there is at most one materializable node and at most one observable node, with all
        # other nodes unexecutable. The asset graph will receive only a single `ExternalAssetNode`
        # representing the asset. This will always be the materializable node if one exists; then
        # the observable node if it exists; then finally the first-encountered unexecutable node.
        for repo_handle, node in materializable_node_pairs:
            asset_nodes_by_key[node.asset_key] = node

        for repo_handle, node in observable_node_pairs:
            if node.asset_key in asset_nodes_by_key:
                current_node = asset_nodes_by_key[node.asset_key]
                asset_nodes_by_key[node.asset_key] = current_node._replace(is_observable=True)
            else:
                asset_nodes_by_key[node.asset_key] = node

        for repo_handle, node in unexecutable_node_pairs:
            if node.asset_key in asset_nodes_by_key:
                continue
            asset_nodes_by_key[node.asset_key] = node

        asset_checks_by_key: Dict[AssetCheckKey, "ExternalAssetCheck"] = {}
        for asset_check in external_asset_checks:
            asset_checks_by_key[asset_check.key] = asset_check

        return cls(
            asset_nodes_by_key,
            asset_checks_by_key,
            repo_handles_by_key=repo_handles_by_key,
        )

    @property
    def asset_nodes(self) -> Sequence["ExternalAssetNode"]:
        return list(self._asset_nodes_by_key.values())

    def get_asset_node(self, asset_key: AssetKey) -> "ExternalAssetNode":
        return self._asset_nodes_by_key[asset_key]

    def has_asset(self, asset_key: AssetKey) -> bool:
        return asset_key in self._asset_nodes_by_key

    @property
    def asset_checks(self) -> Sequence["ExternalAssetCheck"]:
        return list(self._asset_checks_by_key.values())

    def get_asset_check(self, asset_check_key: AssetCheckKey) -> "ExternalAssetCheck":
        return self._asset_checks_by_key[asset_check_key]

    @property
    @cached_method
    def asset_dep_graph(self) -> DependencyGraph[AssetKey]:
        upstream = {node.asset_key: set() for node in self.asset_nodes}
        downstream = {node.asset_key: set() for node in self.asset_nodes}
        for node in self.asset_nodes:
            for dep in node.dependencies:
                upstream[node.asset_key].add(dep.upstream_asset_key)
                downstream[dep.upstream_asset_key].add(node.asset_key)
        return {"upstream": upstream, "downstream": downstream}

    @property
    @cached_method
    def all_asset_keys(self) -> AbstractSet[AssetKey]:
        return {node.asset_key for node in self.asset_nodes}

    @property
    @cached_method
    def materializable_asset_keys(self) -> AbstractSet[AssetKey]:
        return {
            node.asset_key
            for node in self.asset_nodes
            if node.execution_type == AssetExecutionType.MATERIALIZATION
        }

    def is_materializable(self, asset_key: AssetKey) -> bool:
        return self.get_asset_node(asset_key).execution_type == AssetExecutionType.MATERIALIZATION

    @property
    @cached_method
    def observable_asset_keys(self) -> AbstractSet[AssetKey]:
        return {
            node.asset_key
            for node in self.asset_nodes
            # check the separate `is_observable` field because `execution_type` will be
            # `MATERIALIZATION` if there exists a materializable version of the asset
            if node.is_observable
        }

    def is_observable(self, asset_key: AssetKey) -> bool:
        return self.get_asset_node(asset_key).is_observable

    @property
    @cached_method
    def external_asset_keys(self) -> AbstractSet[AssetKey]:
        return {
            node.asset_key
            for node in self.asset_nodes
            if node.execution_type != AssetExecutionType.MATERIALIZATION
        }

    def is_external(self, asset_key: AssetKey) -> bool:
        return self.get_asset_node(asset_key).execution_type != AssetExecutionType.MATERIALIZATION

    @property
    @cached_method
    def executable_asset_keys(self) -> AbstractSet[AssetKey]:
        return {node.asset_key for node in self.asset_nodes if node.is_executable}

    def is_executable(self, asset_key: AssetKey) -> bool:
        return self.get_asset_node(asset_key).execution_type != AssetExecutionType.UNEXECUTABLE

    def asset_keys_for_group(self, group_name: str) -> AbstractSet[AssetKey]:
        return {node.asset_key for node in self.asset_nodes if node.group_name == group_name}

    def asset_keys_for_job(self, job_name: str) -> AbstractSet[AssetKey]:
        return {node.asset_key for node in self.asset_nodes if job_name in node.job_names}

    def get_required_asset_and_check_keys(
        self, asset_or_check_key: AssetKeyOrCheckKey
    ) -> AbstractSet[AssetKeyOrCheckKey]:
        if isinstance(asset_or_check_key, AssetKey):
            atomic_execution_unit_id = self.get_asset_node(
                asset_or_check_key
            ).atomic_execution_unit_id
        else:  # AssetCheckKey
            atomic_execution_unit_id = self.get_asset_check(
                asset_or_check_key
            ).atomic_execution_unit_id
        if atomic_execution_unit_id is None:
            return set()
        else:
            return {
                *(
                    node.asset_key
                    for node in self.asset_nodes
                    if node.atomic_execution_unit_id == atomic_execution_unit_id
                ),
                *(
                    node.key
                    for node in self.asset_checks
                    if node.atomic_execution_unit_id == atomic_execution_unit_id
                ),
            }

    def get_required_multi_asset_keys(self, asset_key: AssetKey) -> AbstractSet[AssetKey]:
        return {
            key
            for key in self.get_required_asset_and_check_keys(asset_key)
            if isinstance(key, AssetKey)
        }

    @property
    @cached_method
    def all_job_names(self) -> AbstractSet[str]:
        return {job_name for node in self.asset_nodes for job_name in node.job_names}

    @property
    @cached_method
    def all_group_names(self) -> AbstractSet[str]:
        return {node.group_name for node in self.asset_nodes if node.group_name}

    def get_partitions_def(self, asset_key: AssetKey) -> Optional[PartitionsDefinition]:
        external_def = self.get_asset_node(asset_key).partitions_def_data
        return external_def.get_partitions_definition() if external_def else None

    def get_partition_mappings(
        self, asset_key: AssetKey
    ) -> Optional[Mapping[AssetKey, PartitionMapping]]:
        return {
            dep.upstream_asset_key: dep.partition_mapping
            for dep in self.get_asset_node(asset_key).dependencies
            if dep.partition_mapping is not None
        }

    def get_group_name(self, asset_key: AssetKey) -> Optional[str]:
        return self.get_asset_node(asset_key).group_name

    def get_freshness_policy(self, asset_key: AssetKey) -> Optional[FreshnessPolicy]:
        return self.get_asset_node(asset_key).freshness_policy

    def get_auto_materialize_policy(self, asset_key: AssetKey) -> Optional[AutoMaterializePolicy]:
        return self.get_asset_node(asset_key).auto_materialize_policy

    def get_auto_observe_interval_minutes(self, asset_key: AssetKey) -> Optional[float]:
        return self.get_asset_node(asset_key).auto_observe_interval_minutes

    def get_backfill_policy(self, asset_key: AssetKey) -> Optional[BackfillPolicy]:
        return self.get_asset_node(asset_key).backfill_policy

    def get_code_version(self, asset_key: AssetKey) -> Optional[str]:
        return self.get_asset_node(asset_key).code_version

    @property
    def repository_handles_by_key(self) -> Mapping[AssetKey, RepositoryHandle]:
        return self._repo_handles_by_key

    def get_repository_handle(self, asset_key: AssetKey) -> RepositoryHandle:
        return self._repo_handles_by_key[asset_key]

    def get_materialization_job_names(self, asset_key: AssetKey) -> Sequence[str]:
        """Returns the names of jobs that materialize this asset."""
        return self.get_asset_node(asset_key).job_names

    def get_materialization_asset_keys_for_job(self, job_name: str) -> Sequence[AssetKey]:
        """Returns asset keys that are targeted for materialization in the given job."""
        return [
            k
            for k in self.materializable_asset_keys
            if job_name in self.get_materialization_job_names(k)
        ]

    def get_implicit_job_name_for_assets(
        self,
        asset_keys: Iterable[AssetKey],
        external_repo: Optional[ExternalRepository],
    ) -> Optional[str]:
        """Returns the name of the asset base job that contains all the given assets, or None if there is no such
        job.

        Note: all asset_keys should be in the same repository.
        """
        if all(self.is_observable(asset_key) for asset_key in asset_keys):
            if external_repo is None:
                check.failed(
                    "external_repo must be passed in when getting job names for observable assets"
                )
            # for observable assets, we need to select the job based on the partitions def
            target_partitions_defs = {
                self.get_partitions_def(asset_key) for asset_key in asset_keys
            }
            check.invariant(len(target_partitions_defs) == 1, "Expected exactly one partitions def")
            target_partitions_def = next(iter(target_partitions_defs))

            # create a mapping from job name to the partitions def of that job
            partitions_def_by_job_name = {}
            for (
                external_partition_set_data
            ) in external_repo.external_repository_data.external_partition_set_datas:
                if external_partition_set_data.external_partitions_data is None:
                    partitions_def = None
                else:
                    partitions_def = external_partition_set_data.external_partitions_data.get_partitions_definition()
                partitions_def_by_job_name[external_partition_set_data.job_name] = partitions_def
            # add any jobs that don't have a partitions def
            for external_job in external_repo.get_all_external_jobs():
                job_name = external_job.external_job_data.name
                if job_name not in partitions_def_by_job_name:
                    partitions_def_by_job_name[job_name] = None
            # find the job that matches the expected partitions definition
            for job_name, external_partitions_def in partitions_def_by_job_name.items():
                asset_keys_for_job = self.asset_keys_for_job(job_name)
                if not job_name.startswith(ASSET_BASE_JOB_PREFIX):
                    continue
                if (
                    # unpartitioned observable assets may be materialized in any job
                    target_partitions_def is None
                    or external_partitions_def == target_partitions_def
                ) and all(asset_key in asset_keys_for_job for asset_key in asset_keys):
                    return job_name
        else:
            for job_name in self.all_job_names:
                asset_keys_for_job = self.asset_keys_for_job(job_name)
                if not job_name.startswith(ASSET_BASE_JOB_PREFIX):
                    continue
                if all(asset_key in self.asset_keys_for_job(job_name) for asset_key in asset_keys):
                    return job_name
        return None

    def split_asset_keys_by_repository(
        self, asset_keys: AbstractSet[AssetKey]
    ) -> Sequence[AbstractSet[AssetKey]]:
        asset_keys_by_repo = defaultdict(set)
        for asset_key in asset_keys:
            repo_handle = self.get_repository_handle(asset_key)
            asset_keys_by_repo[(repo_handle.location_name, repo_handle.repository_name)].add(
                asset_key
            )
        return list(asset_keys_by_repo.values())


def _warn_on_duplicate_nodes(
    node_pairs: Sequence[Tuple[RepositoryHandle, "ExternalAssetNode"]],
    execution_type: AssetExecutionType,
) -> None:
    repo_handles_by_asset_key: DefaultDict[AssetKey, List[RepositoryHandle]] = defaultdict(list)
    for repo_handle, node in node_pairs:
        repo_handles_by_asset_key[node.asset_key].append(repo_handle)

    duplicates = {k: v for k, v in repo_handles_by_asset_key.items() if len(v) > 1}
    duplicate_lines = []
    for asset_key, repo_handles in duplicates.items():
        locations = [repo_handle.code_location_origin.location_name for repo_handle in repo_handles]
        duplicate_lines.append(f"  {asset_key.to_string()}: {locations}")
    duplicate_str = "\n".join(duplicate_lines)
    if duplicates:
        warnings.warn(
            f"Found {execution_type.value} nodes for some asset keys in multiple code locations."
            f" Only one {execution_type.value} node is allowed per asset key. Duplicates:\n {duplicate_str}"
        )



"""This file contains a set of utilities for dealing with test
environments where we want to go back and forth between
abstractions that reside in user process (e.g. definitions and
reconstructables) and abstractions that reside in host processes
(e.g. handles and externals).

These should only be invoked from contexts where we know this
to be the case.
"""


import dagster._check as check
from dagster._core.definitions.reconstruct import ReconstructableJob, ReconstructableRepository
from dagster._core.host_representation import ExternalJob
from dagster._core.host_representation.external_data import (
    external_job_data_from_def,
)
from dagster._core.origin import JobPythonOrigin, RepositoryPythonOrigin


def recon_job_from_origin(origin: JobPythonOrigin) -> ReconstructableJob:
    check.inst_param(origin, "origin", JobPythonOrigin)
    recon_repo = recon_repository_from_origin(origin.repository_origin)
    return recon_repo.get_reconstructable_job(origin.job_name)


def recon_repository_from_origin(origin: RepositoryPythonOrigin) -> "ReconstructableRepository":
    check.inst_param(origin, "origin", RepositoryPythonOrigin)
    return ReconstructableRepository(
        origin.code_pointer,
        origin.container_image,
        origin.executable_path,
        origin.entry_point,
        origin.container_context,
    )


def external_job_from_recon_job(recon_job, op_selection, repository_handle, asset_selection=None):
    if op_selection or asset_selection:
        sub_recon_job = recon_job.get_subset(
            op_selection=op_selection, asset_selection=asset_selection
        )
        job_def = sub_recon_job.get_definition()
    else:
        job_def = recon_job.get_definition()

    return ExternalJob(
        external_job_data_from_def(job_def),
        repository_handle=repository_handle,
    )



from typing import AbstractSet, Any, Callable, Iterable, Mapping, Optional, Set, Tuple, Union

from typing_extensions import TypeAlias

from dagster import _check as check
from dagster._annotations import experimental
from dagster._config import UserConfigSchema
from dagster._core.definitions.asset_check_result import AssetCheckResult
from dagster._core.definitions.asset_check_spec import AssetCheckSpec
from dagster._core.definitions.asset_checks import (
    AssetChecksDefinition,
    AssetChecksDefinitionInputOutputProps,
)
from dagster._core.definitions.asset_dep import CoercibleToAssetDep
from dagster._core.definitions.asset_in import AssetIn
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.events import AssetKey, CoercibleToAssetKey
from dagster._core.definitions.output import Out
from dagster._core.definitions.policy import RetryPolicy
from dagster._core.definitions.resource_annotation import get_resource_args
from dagster._core.definitions.source_asset import SourceAsset
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._core.execution.build_resources import wrap_resources_for_execution

from ..input import In
from .asset_decorator import (
    build_asset_ins,
    get_function_params_without_context_or_config_or_resources,
    make_asset_deps,
)
from .op_decorator import _Op

AssetCheckFunctionReturn: TypeAlias = AssetCheckResult
AssetCheckFunction: TypeAlias = Callable[..., AssetCheckFunctionReturn]


def _build_asset_check_input(
    name: str,
    asset_key: AssetKey,
    fn: Callable[..., Any],
    additional_ins: Mapping[str, AssetIn],
    additional_deps: Optional[AbstractSet[AssetKey]],
) -> Mapping[AssetKey, Tuple[str, In]]:
    fn_params = get_function_params_without_context_or_config_or_resources(fn)

    if asset_key in (additional_deps or []):
        raise DagsterInvalidDefinitionError(
            f"When defining check '{name}', asset '{asset_key.to_user_string()}' was passed to `asset` and `additional_deps`."
            " It can only be passed to one of these parameters."
        )
    if asset_key in [asset_in.key for asset_in in additional_ins.values()]:
        raise DagsterInvalidDefinitionError(
            f"When defining check '{name}', asset '{asset_key.to_user_string()}' was passed to `asset` and `additional_ins`."
            " It can only be passed to one of these parameters."
        )

    fn_param_names = {param.name for param in fn_params}
    for in_name in additional_ins.keys():
        if in_name not in fn_param_names:
            raise DagsterInvalidDefinitionError(
                f"'{in_name}' is specified in 'additional_ins' but isn't a parameter."
            )

    # if all the fn_params are in additional_ins, then we add the prmary asset as a dep
    if len(fn_params) == len(additional_ins):
        all_deps = {*(additional_deps if additional_deps else set()), asset_key}
        all_ins = additional_ins
    # otherwise there should be one extra fn_param, which is the primary asset. Add that as an input
    elif len(fn_params) == len(additional_ins) + 1:
        primary_asset_param_name = next(
            param.name for param in fn_params if param.name not in additional_ins.keys()
        )
        all_ins = {**additional_ins, primary_asset_param_name: AssetIn(asset_key)}
        all_deps = additional_deps
    else:
        param_names_not_in_additional_ins = sorted(
            [f"'{name}'" for name in (fn_param_names - set(additional_ins.keys()))]
        )
        raise DagsterInvalidDefinitionError(
            f"When defining check '{name}', multiple assets provided as parameters:"
            f" [{', '.join(param_names_not_in_additional_ins)}]. These should either match"
            " the target asset or be specified in 'additional_ins'."
        )

    return build_asset_ins(
        fn=fn,
        asset_ins=all_ins,
        deps=all_deps,
    )


@experimental
def asset_check(
    *,
    asset: Union[CoercibleToAssetKey, AssetsDefinition, SourceAsset],
    name: Optional[str] = None,
    description: Optional[str] = None,
    blocking: bool = False,
    additional_ins: Optional[Mapping[str, AssetIn]] = None,
    additional_deps: Optional[Iterable[CoercibleToAssetDep]] = None,
    required_resource_keys: Optional[Set[str]] = None,
    resource_defs: Optional[Mapping[str, object]] = None,
    config_schema: Optional[UserConfigSchema] = None,
    compute_kind: Optional[str] = None,
    op_tags: Optional[Mapping[str, Any]] = None,
    retry_policy: Optional[RetryPolicy] = None,
) -> Callable[[AssetCheckFunction], AssetChecksDefinition]:
    """Create a definition for how to execute an asset check.

    Args:
        asset (Union[AssetKey, Sequence[str], str, AssetsDefinition, SourceAsset]): The
            asset that the check applies to.
        name (Optional[str]): The name of the check. If not specified, the name of the decorated
            function will be used. Checks for the same asset must have unique names.
        description (Optional[str]): The description of the check.
        blocking (bool): When enabled, runs that include this check and any downstream assets that
            depend on `asset` will wait for this check to complete before starting the downstream
            assets. If the check fails with severity `AssetCheckSeverity.ERROR`, then the downstream
            assets won't execute.
        additional_ins (Optional[Mapping[str, AssetIn]]): A mapping from input name to
            information about the input. These inputs will apply to the underlying op that
            executes the check. These should not include the `asset` parameter, which is
            always included as a dependency.
        additional_deps (Optional[Iterable[CoercibleToAssetDep]]): Assets that are upstream
            dependencies, but do not correspond to a parameter of the decorated function. These
            dependencies will apply to the underlying op that executes the check. These should not
            include the `asset` parameter, which is always included as a dependency.
        required_resource_keys (Optional[Set[str]]): A set of keys for resources that are required
            by the function that execute the check. These can alternatively be specified by
            including resource-typed parameters in the function signature.
        config_schema (Optional[ConfigSchema): The configuration schema for the check's underlying
            op. If set, Dagster will check that config provided for the op matches this schema and fail
            if it does not. If not set, Dagster will accept any config provided for the op.
        op_tags (Optional[Dict[str, Any]]): A dictionary of tags for the op that executes the check.
            Frameworks may expect and require certain metadata to be attached to a op. Values that
            are not strings will be json encoded and must meet the criteria that
            `json.loads(json.dumps(value)) == value`.
        compute_kind (Optional[str]): A string to represent the kind of computation that executes
            the check, e.g. "dbt" or "spark".
        retry_policy (Optional[RetryPolicy]): The retry policy for the op that executes the check.


    Produces an :py:class:`AssetChecksDefinition` object.


    Example:
        .. code-block:: python

            from dagster import asset, asset_check, AssetCheckResult

            @asset
            def my_asset() -> None:
                ...

            @asset_check(asset=my_asset, description="Check that my asset has enough rows")
            def my_asset_has_enough_rows() -> AssetCheckResult:
                num_rows = ...
                return AssetCheckResult(passed=num_rows > 5, metadata={"num_rows": num_rows})


    Example with a DataFrame Output:
        .. code-block:: python

            from dagster import asset, asset_check, AssetCheckResult
            from pandas import DataFrame

            @asset
            def my_asset() -> DataFrame:
                ...

            @asset_check(asset=my_asset, description="Check that my asset has enough rows")
            def my_asset_has_enough_rows(my_asset: DataFrame) -> AssetCheckResult:
                num_rows = my_asset.shape[0]
                return AssetCheckResult(passed=num_rows > 5, metadata={"num_rows": num_rows})
    """
    check.opt_set_param(required_resource_keys, "required_resource_keys", of_type=str)
    dict(check.opt_mapping_param(resource_defs, "resource_defs", key_type=str))

    def inner(fn: AssetCheckFunction) -> AssetChecksDefinition:
        check.callable_param(fn, "fn")
        resolved_name = name or fn.__name__
        asset_key = AssetKey.from_coercible_or_definition(asset)

        additional_dep_keys = set([dep.asset_key for dep in make_asset_deps(additional_deps) or []])
        input_tuples_by_asset_key = _build_asset_check_input(
            resolved_name,
            asset_key,
            fn,
            additional_ins=additional_ins or {},
            additional_deps=additional_dep_keys,
        )

        # additional_deps on AssetCheckSpec holds the keys passed to additional_deps and
        # additional_ins. We don't want to include the primary asset key in this set.
        additional_ins_and_deps = input_tuples_by_asset_key.keys() - {asset_key}

        spec = AssetCheckSpec(
            name=resolved_name,
            description=description,
            asset=asset_key,
            additional_deps=additional_ins_and_deps,
            blocking=blocking,
        )

        arg_resource_keys = {arg.name for arg in get_resource_args(fn)}

        check.param_invariant(
            len(required_resource_keys or []) == 0 or len(arg_resource_keys) == 0,
            "Cannot specify resource requirements in both @asset_check decorator and as arguments"
            " to the decorated function",
        )

        resource_defs_keys = set(resource_defs.keys() if resource_defs else [])
        decorator_resource_keys = (required_resource_keys or set()) | resource_defs_keys

        op_required_resource_keys = decorator_resource_keys - arg_resource_keys

        out = Out(dagster_type=None)

        op_def = _Op(
            name=spec.get_python_identifier(),
            ins=dict(input_tuples_by_asset_key.values()),
            out=out,
            # Any resource requirements specified as arguments will be identified as
            # part of the Op definition instantiation
            required_resource_keys=op_required_resource_keys,
            tags={
                **({"kind": compute_kind} if compute_kind else {}),
                **(op_tags or {}),
            },
            config_schema=config_schema,
            retry_policy=retry_policy,
        )(fn)

        checks_def = AssetChecksDefinition(
            node_def=op_def,
            resource_defs=wrap_resources_for_execution(resource_defs),
            specs=[spec],
            input_output_props=AssetChecksDefinitionInputOutputProps(
                asset_keys_by_input_name={
                    input_tuple[0]: asset_key
                    for asset_key, input_tuple in input_tuples_by_asset_key.items()
                },
                asset_check_keys_by_output_name={op_def.output_defs[0].name: spec.key},
            ),
        )

        return checks_def

    return inner
